# This program generates random text, based off of .rtf files (called resume template files).  The aim is to replace the old resume madlibs c++ program with a new program that is easier to modify.  This file should not need to be modified; by changing the template files the text results will be changed.
# Copyright Ryan Beasley and Joanna Lahey 5/25/2012
# Version 8 5/25/2012: Adds Dependent sections.
# Version 7 6/22/2007: The template files now end in ".rtf" because FileGether.exe created files ending in ".dat".
# Version 6 6/22/2007: The ".txt" files are now ".doc" files.  The ".raw" files are now ".txt" files.  This change is being done so that we can continue to use the FileGather.exe program to combine the records for many resumes into a single file.  As a side benefit, the resume files will now open in Word.
# Version 5 6/22/2007: The program now creates ".raw" files also, containing just the filename for the text file and the choices made when creating the file.
# Version 4 6/22/2007: No changes except for the version number (to keep the same as the version of the web-based template generator).
# Version 3 6/22/2007: No changes except for the version number (to keep the same as the version of the web-based template generator).
# Version 2 6/22/2007: Replaced all instances of the text "style" with "template".
# Version 1 6/20/2007

Version = 8

import os
import glob
from random import randrange
from random import random
from random import shuffle
from time import strftime

globalDelayedWrite = []
logging = False

import math
def frange(limit1, limit2 = None, increment = 1.):
  """
  Range function that accepts floats (and integers).

  Usage:
  frange(-2, 2, 0.1)
  frange(10)
  frange(10, increment = 0.5)

  The returned value is an iterator.  Use list(frange(start,...)) for a list.
  """

  if limit2 is None:    limit2, limit1 = limit1, 0.
  else:    limit1 = float(limit1)
  count = int(math.ceil(limit2 - limit1)/increment)
  return (limit1 + n*increment for n in xrange(0,count))


def makeNameArrays(numDifferent, initName, iString):
  nameArray = []
  numDifferentString = str(numDifferent)
  for j in range(numDifferent):
    jString = str(j+1)
    while len(jString)<len(numDifferentString): jString="0"+jString        
    nameArray += [initName+"_"+iString+"_"+jString+"of"+numDifferentString]
  txtArray = map(lambda x: x+'.doc', nameArray)  #note that these are mis-named...see notes for Version 6 above
  savArray = map(lambda x: x+'.sav', nameArray)
  rawArray = map(lambda x: x+'.txt', nameArray)
  return [txtArray, savArray, rawArray]

def createFileNames(name, myTime, numberLength, numDifferent, i=1):
  tempName = name[:-4]+myTime
  iString = str(i)
  while len(iString)<numberLength: iString="0"+iString
  [txtArray, savArray, rawArray] = makeNameArrays(numDifferent, tempName, iString)
  while reduce(lambda x,y: x or os.path.exists(y), [False]+txtArray+savArray+rawArray):
    i+=1
    iString = str(i)
    while len(iString)<numberLength: iString="0"+iString
    [txtArray, savArray, rawArray] = makeNameArrays(numDifferent, tempName, iString)
  return [txtArray, savArray, rawArray]


def createResumes(fileName, numToMake):
  inFile = open(fileName, 'r')
  matchedPair = False
  guiVersion = inFile.readline()
  for line in inFile:
    if '*matchDifferent*' in line or '*matchSame*' in line:
      matchedPair = True
      break
  numDifferent = 1
  if matchedPair:
    while True:
      try: numDifferent = int(raw_input('This template file contains random sections for Matched "pairs."  How many files should be matched? (0 to cancel) '))
      except ValueError:
        print "Please enter a positive integer."
        continue
      if numDifferent < 1: return -1
      break
  myTime = strftime("_%Y-%m-%d-%H-%M-%S")

  for i in range(numToMake):
    [outputFiles, saveChoicesFiles, rawChoicesFiles] = createFileNames(fileName, myTime, len(str(numToMake)), numDifferent, i+1)
    dictionaryMatchSame = {}
    dictionaryMatchDifferent = {}
    for [outputFilename, saveChoicesFilename, rawChoicesFilename] in map(lambda x,y: [x, y[0], y[1]], outputFiles, map(None, saveChoicesFiles, rawChoicesFiles)):
      saveChoicesFile = open(saveChoicesFilename, 'w')
      print >>saveChoicesFile, outputFilename +" is the text file that these choices created"
      print >>saveChoicesFile, fileName+" is the template file being used"
      print >>saveChoicesFile, str(i+1)+" is the index of this text file within a matched set"
      print >>saveChoicesFile, myTime+" is the current time as year, month, day, hour (out of 24), minute, second"
      print >>saveChoicesFile, str(Version)+" is the version of the Python program"
      print >>saveChoicesFile, guiVersion.rstrip('\n')
      print >>saveChoicesFile, str(numDifferent) + " is the number of text files being Matched."
      print >>saveChoicesFile, "Read the following lines in pairs.  The first line is the start tag (from the template file) that required a choice.  The start tag line contains the type of section that required a decision (currently only 'Random' sections require decisions), then the label of this section as shown in the outline in the web-based meta-program, then the number of subsections to choose from, and then any settings for this section (e.g., repeating or matched files).  The second line is the index of the subsection that was randomly chosen.  The indices run from 0 through n-1, inclusive, where n is the number of choices listed in the start tag line.  All of the choices are also stored in the .raw file."
      rawChoicesFile = open(rawChoicesFilename, 'w')
      rawChoicesFile.write(outputFilename)
      outputFile = open(outputFilename, 'w')
      inFile.seek(0)
      inFile.readline() #skip gui version
      retval = recursiveGenerate(inFile, outputFile, saveChoicesFile, rawChoicesFile, '', {}, {}, dictionaryMatchSame, dictionaryMatchDifferent, '', '', '', '', {})
      outputFile.close()
      if (retval<0):
        print >>saveChoicesFile, str(retval)+" is the error code...this template file had a problem"
        rawChoicesFile.write('\t-1')
        print '\nError! Problem with the template file.  Error code '+str(retval)
        saveChoicesFile.close()
        rawChoicesFile.close()
        return retval
      saveChoicesFile.close()
      rawChoicesFile.close()
      print "Done with resume "+outputFilename
  inFile.close()
  return 1

def skipElement(inFile, currentLine):
  try:
    theLine = inFile.readline()
  except StopIteration:
    print "\nError!  The skipElement function ran off the end of the file."
    print "The program chose which subsection to follow for the Random or Dependent section started on this line: "+currentLine
    print "Then skipElement was trying to move down in the template file to that subsection, but did not find it before reaching the end of the file.  Make sure the line above has the correct number of subsections listed after the label (following the second space in the line).  Make sure each subsection has a correct start tag (e.g., *leaf* 1-3-2) and end tag (e.g., *end_leaf* 1-3-2) in the template file."
    raw_input("Press return to continue")
    return -2
  currentLine = theLine
  splitLine = currentLine.split(" ")
  if (len(splitLine)<2):
    print "\nError!  The skipElement function expected a start tag (e.g., *random* 1-3-2 4) but got: "+currentLine
    print "Make sure each start tag is followed by the correct end tag, that each end tag is directly followed by either a start tag or an end tag, and that there are no blank lines in the template file outside of Leaf sections."
    raw_input("Press return to continue")
    return -3
  endTag = "*end_"+splitLine[0][1:]+" "+splitLine[1].rstrip('\n')+" "
  try:
    while not endTag in inFile.readline().rstrip('\n')+' ': pass
  except StopIteration:
    print "\nError!  skipElement ran off the end of the file while looking for the stop tag for: "+currentLine
    print "Make sure that the stop tag exists in the template file (e.g., *end_random* 1-3-2)"
    raw_input("Press return to continue")
    return -4
  return 1

def recursiveGenerate(inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice):
  try:
    theLine = inFile.readline()
  except StopIteration:
    print "\nError!  recursiveGenerate ran off the end of the file.  It expected to find a start tag (e.g., *random* 1-3-1 7).  Make sure the lines (in the template file) that contain start tags for Random and Constant and Dependent sections specify the correct number of subsections listed after the label (following the second space in the line).  Why did the function not find an end tag as the last line in the file?"
    raw_input("Press return to continue")
    return -5
  currentLine = theLine
  if logging: print currentLine
  temp = currentLine.rstrip('\n').split(" ")
  if (len(temp)<2):
    print "\nError!  recursiveGenerate expected a start tag (e.g., *random* 3-2 4) but got: "+currentLine
    print "Make sure the lines (in the template file) that contain start tags for Random and Constant and Dependent sections specify the correct number of subsections listed after the label (following the second space in the line), that each end tag is directly followed by either a start tag or an end tag, and that there are no blank lines in the template file outside of Leaf sections."
    raw_input("Press return to continue")
    return -6
  myText = temp[0]
  myLabel = temp[1]
  myNumChoices = ""
  if (("*random*" in myText) or ("*constant*" in myText) or ("*dependent*" in myText)):
    if (len(temp) < 3):
      print "\nError!  Start tag did not contain number of subsections on this line: "+currentLine
      print "The start tags for Random and Constant and Dependent sections should list the type of the section, then a space, then the label for the section, then a space, then the number of the subsections.  (e.g., *random* 1-1-5-6 8)"
      raw_input("Press return to continue")
      return -7
    temp[2] = temp[2].rstrip("\n")
    if temp[2].isdigit() == False:
      print "\nError!  Number of subsections should be second item following start tag, but on this line that is not a number: "+currentLine
      print "The start tags for Random and Constant and Dependent sections should list the type of the section, then a space, then the label for the section, then a space, then the number of the subsections.  (e.g., *random* 1-1-5-6 8)"
      raw_input("Press return to continue")
      return -8
    myNumChoices = int(temp[2])

  if "*leaf*" in myText:
    return writeLeaf(inFile, outFile, currentLine, myLabel, startString, endString, currentString, currentPlusIntervalString)

  if "*random*" in myText:
    return writeRandom(temp, inFile, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)

  if "*dependent*" in myText:
    return writeDependent(temp, inFile, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)

  if "*constant*" in myText:
    return writeConstant(myNumChoices, myLabel, currentLine, inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)

  print "\nError!  Unrecognized start tag on this line "+currentLine
  print "The recognizable start tags are *leaf*, *random*, *dependent*, and *constant*.  If the line above is not supposed to be a start tag, make sure in the template file it is not directly after a start tag for a Random or Constant or Dependent section, and that it does not directly follow an end tag."
  raw_input("Press return to continue")
  return -9


def writeLeaf(inFile, outFile, currentLine, myLabel, startString, endString, currentString, currentPlusIntervalString):
  try:
    theLine = inFile.readline()
  except StopIteration:
    print "\nError!  The last line in the file is " + currentLine
    print "So there is no *end_leaf* for that tag."
    raw_input("Press return to continue")
    return -10
  firstRun = True
  while not "*end_leaf* " + myLabel in theLine+' ':
    if firstRun:
      firstRun = False
      myLineBreak = ''
    else: myLineBreak = '\n'
    tempString = myLineBreak + theLine.rstrip("\n").replace('%start%',startString).replace('%end%',endString).replace('%current%',currentString).replace('%currentPlusInterval%',currentPlusIntervalString)
    global globalDelayedWrite
    if len(globalDelayedWrite)>0 or '%next%' in tempString: globalDelayedWrite += [tempString]
    else: outFile.write(tempString)
    try:
      theLine = inFile.readline()
    except StopIteration:
      print "\nError!  Could not find *end_leaf* " + myLabel
      print "The program was processing a Leaf section and never found the end tag for that section.  Make sure the end tag is in the file."
      raw_input("Press return to continue")
      return -11
  return 1


def writeConstant(myNumChoices, myLabel, currentLine, inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice):
  for i in range(myNumChoices):
    retval = recursiveGenerate(inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)
    if (retval < 0): return retval
  try:
    theLine = inFile.readline()
  except StopIteration:
    print "\nError!  Ran off the end of the file looking for the *end_constant* corresponding to this start tag "+ currentLine
    print "The program was processing a Constant section and had finished going through the subsections but in doing so it got to the end of the file.  Make sure the end tag is in the correct place and that the Constant section has the correct number of subsections."
    raw_input("Press return to continue")
    return -12
  if ("*end_constant*" + myLabel+" " in theLine):
    print "\nError!  While processing a Constant section, and after processing the subsections, the next line was not *end_constant* corresponding to this start tag "+ currentLine
    print "Make sure the end tag is in the correct location and that the Constant section has the correct number of subsections."
    raw_input("Press return to continue")
    return -13
  return 1


def writeRandom(temp, inFile, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice):
  repeat = False
  repeatSame = False
  repeatNever = False
  repeatNoDoubles = False
  repeatDifferentDouble = False
  repeatDifferentDoublePercentage = 0
  matchSame = False
  matchDifferent = False
  for countSplit in range(3, len(temp)):
    if (temp[countSplit] == "*repeat*") and (len(temp) > countSplit+3):
      repeat = True
      repeatIndicesAreIntegers = True
      try:
        repeatStart = int(temp[countSplit+1])
      except ValueError:
        repeatStart = float(temp[countSplit+1])
        repeatIndicesAreIntegers = False
      try:
        repeatEnd = int(temp[countSplit+2])
      except ValueError:
        repeatEnd = float(temp[countSplit+2])
        repeatIndicesAreIntegers = False
      try:
        repeatInterval = int(temp[countSplit+3])
      except ValueError:
        repeatInterval = float(temp[countSplit+3])
        repeatIndicesAreIntegers = False
    elif temp[countSplit] == "*repeatSame*": repeatSame = True
    elif temp[countSplit] == "*repeatNever*": repeatNever = True
    elif temp[countSplit] == "*repeatNoDoubles*": repeatNoDoubles = True
    elif (temp[countSplit] == "*repeatDifferentDouble*") and (len(temp) > countSplit+1):
      repeatDifferentDouble = True
      repeatDifferentDoublePercentage = float(temp[countSplit+1])
    elif temp[countSplit] == "*matchSame*": matchSame = True
    elif temp[countSplit] == "*matchDifferent*": matchDifferent = True
  if matchDifferent and matchSame:
    print "\nError! This Random start tag: " + currentLine
    print "specifies both Match Different and Match Same, but the two constraints are exclusive.  Fix the stle file by removing one of the two constraints."
    raw_input("Press return")
    return -20
  if repeatSame and repeatNever:
    print "\nError! This Random start tag: " + currentLine
    print "specifies both Repeat Same and Repeat Never, but the two constraints are exclusive.  Fix the stle file by removing one of the two constraints."
    raw_input("Press return")
    return -21
  if repeatSame and repeatDifferentDouble:
    print "\nError! This Random start tag: " + currentLine
    print "specifies both Repeat Same and Repeat Different Double (altering the chance of choosing the same choice again immediately), but the two constraints are exclusive.  Fix the stle file by removing one of the two constraints."
    raw_input("Press return")
    return -22
  if repeatNever and repeatDifferentDouble:
    print "\nError! This Random start tag: " + currentLine
    print "specifies both Repeat Never and Repeat Different Double (altering the chance of choosing the same choice again immediately), but the two constraints are exclusive.  Fix the stle file by removing one of the two constraints."
    raw_input("Press return")
    return -23
  if repeat:
    currentPosition = inFile.tell()
    if repeatIndicesAreIntegers: myRange = range(repeatStart, repeatEnd, repeatInterval)
    else: myRange = frange(repeatStart, repeatEnd, repeatInterval)
    if len(myRange)<1:
      print "\nError! Invalid start/end/interval values for repetition on this start tag: "+ currentLine
      print "Make sure that the interval (the third number after *repeat*) is not zero, and that the start value (the first number after *repeat*) plus some multiple of the interval equals or is past the end value (the second value after *repeat*)."
      return -25
    for myIteration in myRange:
      inFile.seek(currentPosition)
      startString = str(repeatStart)
      endString = str(repeatEnd)
      currentString = str(myIteration)
      currentPlusIntervalString = str(myIteration+repeatInterval)
      retval = enterRandomSection(repeatSame, repeatNever, repeatNoDoubles, repeatDifferentDouble, repeatDifferentDoublePercentage, matchSame, matchDifferent, theIteration+'-'+str(myIteration), myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, inFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)
      if retval < 0: return retval
    if len(globalDelayedWrite)>0: replaceNextString(endString, outFile, myLabel)
  else: retval = enterRandomSection(repeatSame, repeatNever, repeatNoDoubles, repeatDifferentDouble, repeatDifferentDoublePercentage, matchSame, matchDifferent, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, inFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)
  return retval

def writeDependent(temp, inFile, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice):
  # find previous choice in dictionaryLastChoice
  for countSplit in range(3, len(temp)-1):
    if (temp[countSplit] == "*master*"):
      masterLabel = temp[countSplit + 1]
  if masterLabel in dictionaryLastChoice:
    chosenSubelement = dictionaryLastChoice[masterLabel]
  else:
    print "\nError! This dependent section " + myLabel + " depends upon the section labeled " + masterLabel + " which has not yet made a choice.\nThe current line is: " + currentLine
    print "Make sure this dependent section depends upon a random section, and make sure that random section must always be visited before this dependent section.\n"
    raw_input("Press return to continue")
    return -26
  if chosenSubelement >= myNumChoices:
    print "\nError! This dependent section does not have enough subsections.  The master chose element #" + chosenSubelement + " but this dependent section only has #" + myNumChoices + " subsections.\nThe current line is: " + currentLine
    print "\nMake sure this dependent section depends upon the correct random section, make sure that the random section and the dependent section have the same number of subsections."
    raw_input("Press return to continue")
    return -28

  saveChoicesFile.write(currentLine)
  print >>saveChoicesFile, chosenSubelement
  rawChoicesFile.write('\t'+str(chosenSubelement))
  for i in range(chosenSubelement):
    retval = skipElement(inFile, currentLine)
    if (retval < 0): return retval
  retval = recursiveGenerate(inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)
  if (retval < 0): return retval
  try:
    while not "*end_dependent* "+myLabel+" " in inFile.readline(): pass
  except StopIteration:
    print "\nError!  Could not find *end_dependent* for the Dependent section with the label: " + myLabel
    print "The program finished following a subsection for this Dependent section but was unable to find this Dependent section's end tag.  Make sure the end tag is in the file.  Make sure the Random and Constant and Dependent sections have the correct number of subsections."
    raw_input("Press return to continue")
    return -27
  return 1

def intersection(list1, list2):
  ### returns the intersection and then the items in list2 that are not in list 1
  int_dict = {}
  not_int_dict = {}
  list1_dict = {}
  for e in list1: list1_dict[e] = 1
  for e in list2:
    if list1_dict.has_key(e): int_dict[e] = 1
    else: not_int_dict[e] = 1
  return [int_dict.keys(), not_int_dict.keys()]

def getChoiceForRepeatSame(myLabel, dictionaryRepeatSame, freeToChoose, myKey):
  if myLabel in dictionaryRepeatSame:
    if dictionaryRepeatSame[myLabel] in freeToChoose: chosenSubelement = dictionaryRepeatSame[myLabel]
    else:
      print "\nError! Cannot satisfy both Repeat Same and Match Different."
      print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
      print "\nAny given text file was supposed to choose the same choice each time it encountered this Random section (so this random section, or one of its parents must repeat).  All of the matched text files were supposed to choose different choices on the same iteration of the repetition.  The program was not able to satisfy both constraints.  The most likely cause is that not all of the matched files encountered this Random section on the same iteration of a parent Random section."
      print "\nFor example, if the first text file chose the first choice on the first iteration, then the second file did not encounter this Random section (due to a different choice in a Random parent), then the first file did not encounter this Random section in the second iteration, and finally the second file chose the first choice on the second iteration (a valid choice since it has not yet chosen anything and the other file did not choose on this repetition), then on any future repetition if they both encounter this Random section they will not be able to satisfy both constraints."
      print "\nThis error may not always occur because the files may choose differently by chance, or because they choose the same but never encounter this Random section on the same iteration."
      print "To alleviate this problem: remove one of the constraints, add more choices, reduce the number of matched files, or make the parent Random section Match Same so that all matched files encounter this Random section on the same iterations."
      return -14
  else: chosenSubelement = freeToChoose[0]
  return chosenSubelement


def getChoiceForMatchDifferentAndRepeatNever(myLabel, myKey, dictionaryRepeatNever, freeToChoose, myNumChoices):
###if matchDifferent and repeatNever, need to precompute all results to prevent a locking situation (or test for inevitable locking)  but what about nested repeats?
###This is a complicated possibility.  If the results for each text file were determined randomly, just avoiding previous choices (for this file and others), then it is possible to enter a blocking situation (e.g., files a and b choose between results 0,1,2 for three repeats.  'a' chooses 201, 'b' randomly chooses 02 and then has no valid third choice).  Furthermore, if this section is nested within a Repeat section, then it is possible that not all of the text files will be making this choice on the same repeat iteration, complicating the generation of optimal permutations across all text files.
###We have decided to do this the dumb/simple way.  The code will take each text file as it comes and each choice as it comes, obeying the rules for Match Different and Repeat Never.  If it runs into a blocking situation, it will error out.
  if myLabel in dictionaryRepeatNever:
    [alreadyTaken, freeToChoose] = intersection(dictionaryRepeatNever[myLabel], freeToChoose)
    if len(freeToChoose) < 1:
      [alreadyTaken, freeToChoose] = intersection(dictionaryRepeatNever[myLabel], range(myNumChoices))
      if len(freeToChoose) < 1:
        print "\nError! Disobeying Repeat Never, which says that a section should not be chosen more than once in a single text file.  To alleviate this problem: add more choices or reduce the number of repetitions."
        print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
        raw_input("Press return to continue")
        return -24
      else:
        print "\nError!  Failed to obey both Repeat Never and Match Different.  This section or one of its parents repeats, and this section is supposed to always choose a different result than any chosen previously for this text file or any other.  It failed.  To alleviate this problem: add more choices, reduce the number of matched files, or reduce the number of repetitions.  This error may occur even if there exists a set of permutations of the choices that obeys both restrictions."
        print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
        raw_input("Press return to continue")
        return -15
    shuffle(freeToChoose)
    chosenSubelement = freeToChoose[0]
  else: chosenSubelement = freeToChoose[0]
  return chosenSubelement

def getChoiceForMatchDifferentNotRepeatNever(repeatNever, myLabel, repeatDifferentDouble, dictionaryLastChoice, repeatDifferentDoublePercentage, freeToChoose):
  if repeatDifferentDouble and myLabel in dictionaryLastChoice and dictionaryLastChoice[myLabel] in freeToChoose:
    if random()*100. < repeatDifferentDoublePercentage: chosenSubelement = dictionaryLastChoice[myLabel]
    else:
      freeToChoose.remove(dictionaryLastChoice[myLabel])
      if len(freeToChoose)>0: chosenSubelement = freeToChoose[0]
      else:
        print "Warning!  Disobeying specified probability for repeating a choice within a text file due to Match Different, which requires text files to choose differently in a repeating random section on the same iteration.  This will not prevent the text file from being created."
        print "The label for this Random section: "+str(myLabel)
        chosenSubelement = dictionaryLastChoice[myLabel]
  else: chosenSubelement = freeToChoose[0]
  return chosenSubelement

def getChoiceForRepeatNeverNotMatchDifferent(myLabel, myKey, dictionaryRepeatNever, freeToChoose):
  if myLabel in dictionaryRepeatNever:
    [alreadyTaken, freeToChoose] = intersection(dictionaryRepeatNever[myLabel], freeToChoose)
    if len(freeToChoose) < 1:
      print "\nError! Disobeying Repeat Never, which says that a section should not be chosen more than once in a single text file.  To alleviate this problem: add more choices or reduce the number of repetitions."
      print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
      raw_input("Press return to continue")
      return -15
    shuffle(freeToChoose)
    chosenSubelement = freeToChoose[0]
  else: chosenSubelement = freeToChoose[0]
  return chosenSubelement

def getChoiceForDifferentDouble(repeatDifferentDoublePercentage, dictionaryLastChoice, myLabel, freeToChoose):
  if random()*100. < repeatDifferentDoublePercentage: chosenSubelement = dictionaryLastChoice[myLabel]
  else:
    freeToChoose.remove(dictionaryLastChoice[myLabel])
    freeToChoose += [dictionaryLastChoice[myLabel]] # needed in case there is only one choice
    chosenSubelement = freeToChoose[0]
  return chosenSubelement

def getChoiceForMatchSame(repeatSame, myLabel, myKey, dictionaryMatchSame, dictionaryRepeatSame, repeatNever, dictionaryRepeatNever):
    if repeatSame and myLabel in dictionaryRepeatSame and dictionaryMatchSame[myKey] != dictionaryRepeatSame[myLabel]:
      print "\nError! Cannot satisfy both Match Same and Repeat Same."
      print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
      print "\nThis Random section or one of its parents repeats.  This section is supposed to always choose the same result as it has previously chosen, and is supposed to choose the same result as the matched text files chose on the same iteration of the repetition.  The program was not able to satisfy both requirements.  Most likely this random section is within another random section, and that parent random section does not use Match Same.  So this section does not run on the same iterations for all the matched files.  In the first iteration it did run for this file, no previous file had chosen this section, and this file chose differently than the others.  Then in a future iteration, this file and a previous one both ran, putting the two requirements in conflict."
      print "\nTo solve this problem, make the parent repeating section Match Same, or remove one of the two restrictions.  Alternatively, if the current template file is run again there is a chance that the text files will choose similarly and this error will not appear."
      return -16
    if repeatNever and myLabel in dictionaryRepeatNever and dictionaryMatchSame[myKey] in dictionaryRepeatNever[myLabel]:
      print "\nError! Cannot satisfy both Match Same and Repeat Never."
      print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
      print "\nThis Random section or one of its parents repeats.  This section is supposed to always choose the same result as the matched text files chose on the same iteration of the repetition, and this text file is not supposed to contain duplicates. The program was not able to satisfy both requirements. Most likely this random section is within another random section, and that parent random section does not use Match Same.  So this section does not run on the same iterations for all the matched files.  This text file made a choice in an iteration that no previous file chose during.  Then in a later iteration, the previous files made that same choice and now this file cannot satisfy both requirements."
      print "\nTo solve this problem, make the parent repeating section Match Same, or remove one of the two restrictions.  Alternatively, if the current template file is run again there may be a chance that the text files will choose similarly and this error will not appear."
      return -17
    chosenSubelement = dictionaryMatchSame[myKey]
    return chosenSubelement
  
def getChosenSubElement(repeatSame, repeatNever, repeatNoDoubles, repeatDifferentDouble, repeatDifferentDoublePercentage, matchSame, matchDifferent, theIteration, myNumChoices, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, dictionaryLastChoice):
  myKey = myLabel+'-iter'+theIteration
  if matchDifferent and myKey in dictionaryMatchDifferent:
    [alreadyTaken, freeToChoose] = intersection(dictionaryMatchDifferent[myKey], range(myNumChoices))
    if len(freeToChoose) < 1:
      print "\nError! Disobeying Match Different.  Not enough choices."
      print "The label for this Random section: " + str(myLabel) + ". The 'key' which contains the label and also a concatenated list of the iterations for any ongoing repetitions: " + str(myKey)
      print "\nThis random section or one of its parents repeats.  This section is not supposed to have the same result as any of the text files it is matched with for the same iteration in the repetition.  The program was not able to satisfy that requirement.  Add more choices, reduce the number of matched files, or reduce the number of repetitions (check for nested repeating sections)."
      raw_input("Press return to continue")
      return [-19, -1]
  else: freeToChoose = range(myNumChoices)
  shuffle(freeToChoose)

  if matchSame and myKey in dictionaryMatchSame: chosenSubelement = getChoiceForMatchSame(repeatSame, myLabel, myKey, dictionaryMatchSame, dictionaryRepeatSame, repeatNever, dictionaryRepeatNever)
  elif repeatSame: chosenSubelement = getChoiceForRepeatSame(myLabel, dictionaryRepeatSame, freeToChoose, myKey)
  elif matchDifferent and repeatNever: chosenSubelement = getChoiceForMatchDifferentAndRepeatNever(myLabel, myKey, dictionaryRepeatNever, freeToChoose, myNumChoices)
  elif matchDifferent and myKey in dictionaryMatchDifferent: chosenSubelement = getChoiceForMatchDifferentNotRepeatNever(repeatNever, myLabel, repeatDifferentDouble, dictionaryLastChoice, repeatDifferentDoublePercentage, freeToChoose)
  elif repeatNever: chosenSubelement = getChoiceForRepeatNeverNotMatchDifferent(myLabel, myKey, dictionaryRepeatNever, freeToChoose)
  elif repeatDifferentDouble and myLabel in dictionaryLastChoice: chosenSubelement = getChoiceForDifferentDouble(repeatDifferentDoublePercentage, dictionaryLastChoice, myLabel, freeToChoose)
  else: chosenSubelement = freeToChoose[0]
  if chosenSubelement < 0: return [chosenSubelement, -1]

  if repeatSame: dictionaryRepeatSame[myLabel] = chosenSubelement
  if matchSame: dictionaryMatchSame[myKey] = chosenSubelement
  if matchDifferent:
    if myKey in dictionaryMatchDifferent: dictionaryMatchDifferent[myKey] += [chosenSubelement]
    else: dictionaryMatchDifferent[myKey] = [chosenSubelement]
  if repeatNever:
    if myLabel in dictionaryRepeatNever: dictionaryRepeatNever[myLabel] += [chosenSubelement]
    else: dictionaryRepeatNever[myLabel] = [chosenSubelement]
      
  if myLabel in dictionaryLastChoice and chosenSubelement == dictionaryLastChoice[myLabel]: sameChoiceAsLastTime = True
  else: sameChoiceAsLastTime = False
  dictionaryLastChoice[myLabel] = chosenSubelement
  return [chosenSubelement, sameChoiceAsLastTime]



def enterRandomSection(repeatSame, repeatNever, repeatNoDoubles, repeatDifferentDouble, repeatDifferentDoublePercentage, matchSame, matchDifferent, theIteration, myNumChoices, currentLine, saveChoicesFile, rawChoicesFile, inFile, outFile, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice):
  [chosenSubelement, sameChoiceAsLastTime] = getChosenSubElement(repeatSame, repeatNever, repeatNoDoubles, repeatDifferentDouble, repeatDifferentDoublePercentage, matchSame, matchDifferent, theIteration, myNumChoices, myLabel, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, dictionaryLastChoice)
  if chosenSubelement < 0: return chosenSubelement
  if len(globalDelayedWrite)>0 and not sameChoiceAsLastTime: replaceNextString(currentString, outFile, myLabel)
  saveChoicesFile.write(currentLine)
  print >>saveChoicesFile, chosenSubelement
  rawChoicesFile.write('\t'+str(chosenSubelement))
  for i in range(chosenSubelement):
    retval = skipElement(inFile, currentLine)
    if (retval < 0): return retval
  if not repeatNoDoubles or not sameChoiceAsLastTime:
    retval = recursiveGenerate(inFile, outFile, saveChoicesFile, rawChoicesFile, theIteration, dictionaryRepeatSame, dictionaryRepeatNever, dictionaryMatchSame, dictionaryMatchDifferent, startString, endString, currentString, currentPlusIntervalString, dictionaryLastChoice)
    if (retval < 0): return retval
  try:
    while not "*end_random* "+myLabel+" " in inFile.readline(): pass
  except StopIteration:
    print "\nError!  Could not find *end_random* for the Random section with the label: " + myLabel
    print "The program finished following a subsection for this Random section but was unable to find this Random section's end tag.  Make sure the end tag is in the file.  Make sure the Random and Constant sections have the correct number of subsections."
    raw_input("Press return to continue")
    return -18
  return 1

def replaceNextString(currentString, outFile, myLabel):
    global globalDelayedWrite
    tempList = []
    readyToPrint = True
    for line in globalDelayedWrite:
        if '%next%'+myLabel+'%' in line: replacedLine = line.replace('%next%'+myLabel+'%', currentString)
        else: replacedLine = line
        if '%next%' in replacedLine: readyToPrint = False
        tempList += [replacedLine]
    if readyToPrint:
        for line in tempList: outFile.write(line)
        globalDelayedWrite = []
    else: globalDelayedWrite = tempList

while True:
    print "\n\nResumeRandomizer program"
    templateFileNames = glob.glob('*.rtf')
    templateFileNames = filter(os.path.isfile, templateFileNames)
    if templateFileNames is None or len(templateFileNames)<1:
        print "No .rtf file available...where are the resume template files?  They should be in the same folder as this program."
        raw_input("Press return to quit")
        break
    for i in range(len(templateFileNames)): print str(i+1) + ") " + templateFileNames[i]
    try:
        whichTemplate = int(raw_input('Which template?  (0 to quit) '))
    except ValueError:
        print "Please enter an integer between 0 and "+str(len(templateFileNames))
        continue
    if (whichTemplate < 1):
        print "Bye"
        break
    if (whichTemplate > len(templateFileNames)):
        raw_input("That number is too large.  Press return")
        continue
    whichTemplate-=1
    print "Using template "+ templateFileNames[whichTemplate]
    try:
        howMany = int(raw_input('How many? (0 to cancel) '))
    except ValueError:
        print "Please enter a positive integer."
        continue
    if (howMany < 1): print "Canceled"
    else: createResumes(templateFileNames[whichTemplate], howMany)
